from __future__ import annotations

import json
import logging
from asyncio import Queue
from datetime import datetime
from pathlib import Path
from typing import List

from rich.console import RenderableType
from rich.containers import Renderables
from rich.logging import RichHandler
from rich.text import Text


class CustomHandler(logging.Handler):
    def __init__(self, callback):
        """Set up a handler instance, record the callback function."""
        super().__init__()
        self._callback = callback

    def prepare(self, record):
        self.format(record)
        record_dict = record.__dict__
        record_dict["type"] = "log"
        return json.dumps(record_dict)

    def emit(self, record):
        try:
            self._callback(self.prepare(record))
        except Exception:
            self.handleError(record)


class DirectableRichHandler(RichHandler):
    def __init__(self, queue: Queue, **kwargs):
        super().__init__(**kwargs)
        self._queue = queue
        self.redirect = False
        self._last_time = None

    def get_log_row(self, record, message_renderable) -> list:
        row: List[RenderableType] = []
        renderables = [message_renderable]
        path = Path(record.pathname).name
        level = self.get_level_text(record)
        time_format = None if self.formatter is None else self.formatter.datefmt
        log_time = datetime.fromtimestamp(record.created) or datetime.now()
        time_format = time_format or self._log_render.time_format
        link_path = record.pathname if self.enable_link_path else None
        if callable(time_format):
            log_time_display = time_format(log_time)
        else:
            log_time_display = Text(log_time.strftime(time_format))
        if log_time_display == self._last_time:
            row.append(Text(" " * len(log_time_display)))
        else:
            row.append(log_time_display)
            self._last_time = log_time_display
        row.append(level)

        row.append(Renderables(renderables))
        if path:
            path_text = Text()
            path_text.append(
                path, style=f"link file://{link_path}" if link_path else ""
            )
            if record.lineno:
                path_text.append(":")
                path_text.append(
                    f"{record.lineno}",
                    style=f"link file://{link_path}#{record.lineno}"
                    if link_path
                    else "",
                )
            row.append(path_text)
        return row

    def emit(self, record):
        try:
            if self.redirect:
                message = self.format(record)
                message_renderable = self.render_message(record, message)
                rendered_log = self.render(
                    record=record,
                    traceback=None,
                    message_renderable=message_renderable,
                )
                rendered_log_row = self.get_log_row(
                    record=record, message_renderable=message_renderable
                )
                self._queue.put((rendered_log_row, rendered_log))
            else:
                super().emit(record)
        except Exception:
            self.handleError(record)



from __future__ import annotations

# import asyncio
# import contextlib
import copy
import logging
import string
import threading
from pathlib import Path
from queue import Queue
from typing import Callable, Dict, List, NamedTuple, OrderedDict, TypeVar, Union
from urllib.parse import urlparse

import requests
from pydantic import BaseModel, ValidationError
from rich.box import MINIMAL, SQUARE
from rich.logging import RichHandler
from rich.panel import Panel
from textual import events
from textual.app import App
from textual.keys import Keys
from textual.reactive import Reactive
from textual.views import DockView
from textual.widget import Widget
from textual.widgets import ScrollView

from murfey.client.analyser import Analyser
from murfey.client.context import SPAContext, TomographyContext
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.client.rsync import RSyncer, RSyncerUpdate, TransferResult
from murfey.client.tui.forms import TUIFormValue
from murfey.client.tui.status_bar import StatusBar

log = logging.getLogger("murfey.tui.app")

ReactiveType = TypeVar("ReactiveType")

# @contextlib.asynccontextmanager
# async def async_lock(lock):
#     loop = asyncio.get_event_loop()
#     await loop.run_in_executor(_pool, lock.acquire)
#     try:
#         yield
#     finally:
#         lock.release()


class InputResponse(NamedTuple):
    question: str
    allowed_responses: List[str] | None = None
    default: str = ""
    callback: Callable | None = None
    key_change_callback: Callable | None = None
    kwargs: dict | None = None
    form: OrderedDict[str, TUIFormValue] | None = None
    model: BaseModel | None = None


class InfoWidget(Widget):
    text: Reactive[str] = Reactive("")

    def __init__(self, text: str, **kwargs):
        super().__init__(**kwargs)
        self.text = text

    def render(self) -> Panel:
        return Panel(self.text, style=("on dark_magenta"), box=SQUARE)

    def _key_change(self, input_char: str | None):
        if input_char is None:
            self.text = self.text[:-1]
            return
        self.text += input_char


class HoverVisit(Widget):
    mouse_over = Reactive(False)
    lock: bool | None = None

    def __init__(self, text: str, **kwargs):
        super().__init__(**kwargs)
        self._text = text

    def render(self) -> Panel:
        if self.lock is None:
            return Panel(
                self._text,
                style=("on purple4" if self.mouse_over else "on medium_purple3"),
                box=SQUARE,
            )
        return Panel(
            self._text,
            style=("on purple4" if self.lock else "on bright_black"),
            box=SQUARE,
        )

    def on_enter(self) -> None:
        self.mouse_over = True

    def on_leave(self) -> None:
        self.mouse_over = False

    def on_click(self) -> None:
        if self.lock is None:
            self.lock = True
            if isinstance(self.app, MurfeyTUI):
                for h in self.app.hovers:
                    if isinstance(h, HoverVisit) and h != self:
                        h.lock = False
                        h.refresh()
                self.app.input_box.lock = False
                self.app._visit = self._text
                self.app._environment.visit = self._text
                machine_data = requests.get(
                    f"{self.app._environment.url.geturl()}/machine/"
                ).json()
                _default = ""
                if self.app._default_destination:
                    if (
                        self.app._environment.processing_only_mode
                        and self.app._environment.source
                    ):
                        _default = str(self.app._environment.source.resolve()) or str(
                            Path.cwd()
                        )
                    elif machine_data.get("data_directories"):
                        for data_dir in machine_data["data_directories"].keys():
                            if (
                                self.app._environment.source
                                and self.app._environment.source.resolve()
                                == Path(data_dir)
                            ):
                                _default = (
                                    self.app._default_destination + f"/{self._text}"
                                )
                                if self.app.analyser:
                                    self.app.analyser._role = machine_data[
                                        "data_directories"
                                    ][data_dir]
                                break
                            elif self.app._environment.source:
                                try:
                                    mid_path = self.app._environment.source.resolve().relative_to(
                                        data_dir
                                    )
                                    if (
                                        machine_data["data_directories"][data_dir]
                                        == "detector"
                                    ):
                                        suggested_path_response = requests.post(
                                            url=f"{str(self.app._url.geturl())}/visits/{self._text}/suggested_path",
                                            json={
                                                "base_path": f"{self.app._default_destination}/{self._text}/{mid_path.parent}/raw"
                                            },
                                        )
                                        _default = suggested_path_response.json().get(
                                            "suggested_path"
                                        )
                                    else:
                                        _default = f"{self.app._default_destination}/{self._text}/{mid_path}"
                                    if self.app.analyser:
                                        self.app.analyser._role = machine_data[
                                            "data_directories"
                                        ][data_dir]
                                    break
                                except (ValueError, KeyError):
                                    _default = ""
                        else:
                            _default = ""
                    else:
                        _default = self.app._default_destination + f"/{self._text}"
                else:
                    _default = "unknown"
                if self.app._environment.processing_only_mode:
                    self.app._start_rsyncer(_default)
                else:
                    self.app._queues["input"].put_nowait(
                        InputResponse(
                            question="Transfer to: ",
                            default=_default,
                            callback=self.app._start_rsyncer,
                        )
                    )


class QuickPrompt:
    def __init__(self, text: str, options: List[str]):
        self._text = text
        self._options = options
        self.warn = False

    def __repr__(self):
        return repr(self._text)

    def __str__(self):
        return self._text

    def __iter__(self):
        return iter(self._options)

    def __bool__(self):
        return bool(self._text)


def validate_form(form: dict, model: BaseModel) -> dict:
    try:
        validated = model(**form)
        return validated.dict()
    except (AttributeError, ValidationError) as e:
        log.debug(e)
        return {}


class InputBox(Widget):
    input_text: Union[Reactive[str], str] = Reactive("")
    prompt: str | QuickPrompt | None = ""
    mouse_over = Reactive(False)
    can_focus = True
    lock: bool = True
    current_callback: Callable | None = None
    key_change_callback: Callable | None = None
    _question: str = ""
    _form: Reactive[OrderedDict] = Reactive(OrderedDict({}))

    def __init__(self, app, queue: Queue | None = None):
        self._app_reference = app
        self._queue: Queue = queue or Queue()
        # self._form: dict = {}
        self._line = 0
        self._form_keys: List[str] = []
        self._unanswered_message = False
        self._model: BaseModel | None = None
        super().__init__()

    @property
    def _num_lines(self):
        return len(self._form.keys())

    def render(self) -> Panel:
        if not self._queue.empty() and not self.prompt and not self.input_text:
            msg = self._queue.get_nowait()
            if msg is not None:
                self._unanswered_message = True
            self.input_text = ""
            if msg.form:
                self._form = msg.form
                self._model = msg.model
            if msg.allowed_responses:
                self.prompt = QuickPrompt(msg.question, msg.allowed_responses)
                if msg.callback:
                    self.current_callback = msg.callback
            else:
                self._question = msg.question
                self.input_text = msg.question + msg.default
                if msg.callback:
                    self.current_callback = msg.callback
        if isinstance(self.prompt, QuickPrompt):
            panel_msg = (
                f"{self.prompt}: [[red]{'/'.join(self.prompt)}[/red]] {self.input_text}"
                if self.prompt.warn
                else f"{self.prompt}: [[white]{'/'.join(self.prompt)}[/white]] {self.input_text}"
            )
        elif self._form:
            self._line = 1
            self._form_keys = list(self._form.keys())
            panel_msg = f"{self.input_text}\n" + "\n".join(
                f"[cyan]{key}[/cyan]: {self._form[key]}[blink]\u275a[/blink]"
                if i == self._line - 1
                else f"[cyan]{key}[/cyan]: {self._form[key]}"
                for i, key in enumerate(self._form_keys)
            )
        else:
            panel_msg = f"[white]❯[/white] {self.input_text}[blink]\u275a[/blink]"
        return Panel(
            panel_msg,
            style=(
                "on blue"
                if self.mouse_over and not self._unanswered_message
                else "on deep_pink4"
                if self.mouse_over and self._unanswered_message
                else "on red"
                if self._unanswered_message
                else ""
            ),
            box=SQUARE,
        )

    def on_mount(self):
        self.set_interval(2, self.tick)

    def tick(self):
        self.refresh()

    def set_input_text(self, input_text: str) -> None:
        self.input_text = input_text

    async def on_enter(self) -> None:
        if not self.lock:
            self.mouse_over = True
            await self.focus()

    async def on_leave(self) -> None:
        if not self.lock:
            self.mouse_over = False
            await self._app_reference.set_focus(None)

    async def on_key(self, key: events.Key) -> None:
        if key.key == Keys.ControlH and (
            self.input_text != self._question or self._line
        ):
            if self._line == 0:
                self.input_text = self.input_text[:-1]
                if self.key_change_callback:
                    self.key_change_callback(None)
            else:
                k = self._form_keys[self._line - 1]
                # set self._form rather than accessing by key in order to make use of reactivity
                self._form = OrderedDict(
                    {
                        _k: TUIFormValue(self._form[_k].data[:-1])
                        if _k == k
                        else self._form[_k]
                        for _k in self._form_keys
                    }
                )
            key.stop()
        elif key.key == Keys.Delete:
            self._form = OrderedDict({})
            self.input_text = ""
            key.stop()
        elif key.key == Keys.Down:
            new_line = self._line + 1
            if new_line <= self._num_lines:
                self._line = new_line
            key.stop()
        elif key.key == Keys.Up:
            new_line = self._line - 1
            if new_line >= 0:
                self._line = new_line
            key.stop()
        elif key.key in string.printable:
            if self._line == 0:
                self.input_text += key.key
                if self.key_change_callback:
                    self.key_change_callback(key.key)
            else:
                k = self._form_keys[self._line - 1]
                # set self._form rather than accessing by key in order to make use of reactivity
                self._form = OrderedDict(
                    {
                        _k: TUIFormValue(self._form[_k].data + key.key)
                        if _k == k
                        else self._form[_k]
                        for _k in self._form_keys
                    }
                )
            key.stop()
        elif key.key == Keys.Enter and self.prompt:
            if self.input_text not in self.prompt and isinstance(
                self.prompt, QuickPrompt
            ):
                self.prompt.warn = True
            else:
                self.prompt = None
                if self.current_callback:
                    self.current_callback(
                        self.input_text.replace(self._question, "", 1)
                    )
                    self.current_callback = None
                self.input_text = ""
                self._unanswered_message = False
            key.stop()
        elif key.key == Keys.Enter and self.current_callback:
            if self._form:
                if validated_form := validate_form(
                    {k: v.data for k, v in self._form.items()}, self._model
                ):
                    self.current_callback(validated_form)
                    self._form = OrderedDict({})
                    self._form_keys = []
                else:
                    return
            else:
                self.current_callback(self.input_text.replace(self._question, "", 1))
            self.current_callback = None
            self.input_text = ""
            self._unanswered_message = False
            key.stop()
        elif key.key == Keys.Enter:
            self.input_text = ""
            self._unanswered_message = False
            if self._form:
                self._form = OrderedDict({})
                self._form_keys = []
            key.stop()


class LogBook(Widget):
    def __init__(self, queue, *args, **kwargs):
        self._queue = queue
        self._next_log = None
        self._logs = None
        self._log_cache = []
        self._handler = RichHandler(enable_link_path=False)
        super().__init__(*args, **kwargs)

    def on_mount(self):
        self.set_interval(0.5, self.tick)

    def _load_from_queue(self) -> bool:
        if not self._queue.empty():
            num_logs = 0
            self._next_log = []
            while not self._queue.empty() and num_logs < 10:
                msg = self._queue.get_nowait()
                self._next_log.append(msg)
                num_logs += 1
            return True
        return False

    def render(self) -> Panel:
        panel_msg = self._logs or ""
        return Panel(
            panel_msg,
            box=MINIMAL,
        )

    async def tick(self):
        loaded = self._load_from_queue()
        if loaded:
            if self._logs is None:
                self._logs = self._next_log[0][1]
                for nl in self._next_log[1:]:
                    self._log_cache.append(nl)
                    self._logs.add_row(*nl[0])
            else:
                for nl in self._next_log:
                    self._log_cache.append(nl)
                    self._logs.add_row(*nl[0])
            if len(self._log_cache) > 50:
                self._logs = self._log_cache[-50][1]
                curr_log_cache = copy.deepcopy(self._log_cache)
                for r in curr_log_cache[-49:]:
                    self._logs.add_row(*r[0])
                self._log_cache = curr_log_cache[-50:]
                del curr_log_cache
            self.refresh()


class DCParametersTomo(BaseModel):
    voltage: float
    pixel_size_on_image: str
    experiment_type: str
    image_size_x: int
    image_size_y: int
    tilt: int
    acquisition_software: str
    dose_per_frame: float


class MurfeyTUI(App):
    input_box: InputBox
    log_book: ScrollView
    hover: List[str]
    visits: List[str]
    rsync_process: RSyncer | None = None
    analyser: Analyser | None = None

    def __init__(
        self,
        environment: MurfeyInstanceEnvironment | None = None,
        visits: List[str] | None = None,
        queues: Dict[str, Queue] | None = None,
        status_bar: StatusBar | None = None,
        dummy_dc: bool = True,
        do_transfer: bool = True,
        rsync_process: RSyncer | None = None,
        analyser: Analyser | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._environment = environment or MurfeyInstanceEnvironment(
            urlparse("http://localhost:8000")
        )
        self._source = self._environment.source or Path(".")
        self._url = self._environment.url
        self._default_destination = self._environment.default_destination
        self._watcher = self._environment.watcher
        self.visits = visits or []
        self._queues = queues or {}
        self._statusbar = status_bar or StatusBar()
        self._request_destinations = False
        self._register_dc: bool | None = None
        self._tmp_responses: List[dict] = []
        self._visit = ""
        self._dc_metadata: dict = {}
        self._dummy_dc = dummy_dc
        self._do_transfer = do_transfer
        self.rsync_process = rsync_process
        self.analyser = analyser
        self._data_collection_form_complete = False
        self._info_widget = InfoWidget("Welcome to Murfey :microscope: \n")

    @property
    def role(self) -> str:
        if self.analyser:
            return self.analyser._role
        return ""

    def _start_rsyncer(self, destination: str):
        new_rsyncer = False
        if self._environment:
            self._environment.default_destination = destination
        if not self.rsync_process:
            self.rsync_process = RSyncer(
                self._source,
                basepath_remote=Path(destination),
                server_url=self._url,
                local=self._environment.demo,
                status_bar=self._statusbar,
                do_transfer=self._do_transfer,
            )
            new_rsyncer = True
        else:
            if self._environment.demo:
                _remote = destination
            else:
                _remote = f"{self._url.hostname}::{destination}"
            self.rsync_process._remote = _remote
            self.thread = threading.Thread(
                name=f"RSync {self._source.absolute()}:{_remote}",
                target=self.rsync_process._process,
            )

        self._info_widget.text += f"{self._source.resolve()} \u2192 {destination} \n"

        def rsync_result(update: RSyncerUpdate):
            if not self.rsync_process:
                raise ValueError("TUI rsync process does not exist")
            if update.outcome is TransferResult.SUCCESS:
                # log.info(
                #     f"File {str(update.file_path)!r} successfully transferred ({update.file_size} bytes)"
                # )
                pass
            else:
                log.warning(f"Failed to transfer file {str(update.file_path)!r}")
                self.rsync_process.enqueue(update.file_path)

        if self.rsync_process:
            self.rsync_process.subscribe(rsync_result)
            self.rsync_process.start()
            new_analyser = False
            if not self.analyser:
                self.analyser = Analyser(
                    self._source,
                    environment=self._environment if not self._dummy_dc else None,
                )
                new_analyser = True
            if self._watcher:
                if new_rsyncer:
                    self._watcher.subscribe(self.rsync_process.enqueue)
                if new_analyser:
                    self._watcher.subscribe(self.analyser.enqueue)
            self.analyser.subscribe(self._data_collection_form)
            self.analyser.start()

    def _set_register_dc(self, response: str):
        if response == "y":
            self._register_dc = True
            for r in self._tmp_responses:
                self._queues["input"].put_nowait(
                    InputResponse(
                        question="Data collection parameters:",
                        form=r.get("form", OrderedDict({})),
                        model=DCParametersTomo
                        if self.analyser
                        and isinstance(self.analyser._context, TomographyContext)
                        else None,
                        callback=self.app._start_dc,
                    )
                )
                self._dc_metadata = r.get("form", OrderedDict({}))
        elif response == "n":
            self._register_dc = False
        self._tmp_responses = []

    def _data_collection_form(self, response: dict):
        if self._data_collection_form_complete:
            return
        if self._register_dc and response.get("form"):
            self._queues["input"].put_nowait(
                InputResponse(
                    question="Data collection parameters:", form=response["form"]
                )
            )
            self._data_collection_form_complete = True
        elif response.get("allowed_responses"):
            self._queues["input"].put_nowait(
                InputResponse(
                    question="Would you like to start a data collection?",
                    allowed_responses=response["allowed_responses"],
                    callback=self._set_register_dc,
                )
            )
        elif self._register_dc is None:
            self._tmp_responses.append(response)
            self._data_collection_form_complete = True

    def _start_dc(self, json):
        if self._dummy_dc:
            return
        self._environment.data_collection_parameters = json
        self._info_widget.text += "\n".join(f"{k}: {v}" for k, v in json.items()) + "\n"
        if isinstance(self.analyser._context, TomographyContext):
            self._environment.listeners["data_collection_group_id"] = {
                self.analyser._context._flush_data_collections
            }
            self._environment.listeners["data_collection_ids"] = {
                self.analyser._context._flush_processing_job
            }
            self._environment.listeners["autoproc_program_ids"] = {
                self.analyser._context._flush_preprocess
            }
            self._environment.listeners["motion_corrected_movies"] = {
                self.analyser._context._check_for_alignment
            }
            url = f"{str(self._url.geturl())}/visits/{str(self._visit)}/register_data_collection_group"
            dcg_data = {"experiment_type": "tomo", "experiment_type_id": 36}
            requests.post(url, json=dcg_data)
        elif isinstance(self.analyser._context, SPAContext):
            url = f"{str(self._url.geturl())}/visits/{str(self._visit)}/register_data_collection_group"
            dcg_data = {"experiment_type": "single particle", "experiment_type_id": 37}
            requests.post(url, json=dcg_data)
            url = f"{str(self._url.geturl())}/visits/{str(self._visit)}/start_data_collection"
            requests.post(url, json=json)

    def _update_info(self, new_text: str):
        self._info_widget.text = new_text

    def _set_request_destination(self, response: str):
        if response == "y":
            self._request_destinations = True

    async def on_load(self, event):
        await self.bind("q", "quit", show=True)

    async def on_mount(self) -> None:
        self.input_box = InputBox(self, queue=self._queues.get("input"))
        self.log_book = LogBook(self._queues["logs"])
        # self._statusbar = StatusBar()
        self.hovers = (
            [HoverVisit(v) for v in self.visits]
            if len(self.visits)
            else [HoverVisit("No ongoing visits found")]
        )

        grid = await self.view.dock_grid(edge="left")

        grid.add_column(fraction=1, name="left")
        grid.add_column(fraction=1, name="right")
        grid.add_row(fraction=1, name="top")
        grid.add_row(fraction=1, name="middle")
        grid.add_row(fraction=1, name="bottom")

        grid.add_areas(
            area1="left,top",
            area2="right,top-start|bottom-end",
            # area3="left,middle",
            area3="left,middle-start|bottom-end",
        )

        sub_view = DockView()
        await sub_view.dock(*self.hovers, edge="top")

        info_sub_view = DockView()
        await info_sub_view.dock(self.input_box, self._info_widget, edge="top")

        grid.place(
            area1=sub_view,
            area2=self.log_book,
            # area3=self._statusbar,
            # area4=self.input_box,
            area3=info_sub_view,
        )

    async def action_quit(self) -> None:
        if self.rsync_process:
            self.rsync_process.stop()
        if self.analyser:
            self.analyser.stop()
        await self.shutdown()



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional

from pydantic import BaseModel


class Visit(BaseModel):
    start: datetime
    end: datetime
    session_id: int
    name: str
    beamline: str
    proposal_title: str

    def __repr__(self) -> str:
        return (
            "Visit("
            f"start='{self.start:%Y-%m-%d %H:%M}', "
            f"end='{self.end:%Y-%m-%d %H:%M}', "
            f"session_id='{self.session_id!r}',"
            f"name={self.name!r}, "
            f"beamline={self.beamline!r}, "
            f"proposal_title={self.proposal_title!r}"
            ")"
        )


class ContextInfo(BaseModel):
    experiment_type: str
    acquisition_software: str


class File(BaseModel):
    name: str
    description: str
    size: int
    timestamp: float


class ProcessFile(BaseModel):
    path: str
    description: str
    size: int
    timestamp: float
    processing_job: int
    data_collection_id: int
    image_number: int
    mc_uuid: int
    autoproc_program_id: int
    pixel_size: float
    gain_ref: Optional[str] = None


class TiltSeries(BaseModel):
    name: str
    file_tilt_list: str
    dcid: int
    processing_job: int
    autoproc_program_id: int
    motion_corrected_path: str
    movie_id: int


class SuggestedPathParameters(BaseModel):
    base_path: Path


class DCGroupParameters(BaseModel):
    experiment_type: str
    experiment_type_id: int


class DCParameters(BaseModel):
    voltage: float
    pixel_size_on_image: str
    experiment_type: str
    image_size_x: int
    image_size_y: int
    tilt: int
    file_extension: str
    acquisition_software: str
    image_directory: str
    tag: str


class ProcessingJobParameters(BaseModel):
    tag: str
    recipe: str


class RegistrationMessage(BaseModel):
    registration: str
    params: Optional[Dict[str, Any]] = None



from __future__ import annotations

import logging
import os
import time
from pathlib import Path
from typing import NamedTuple, Optional

import murfey.util
from murfey.client.tui.status_bar import StatusBar

log = logging.getLogger("murfey.client.watchdir")


class _FileInfo(NamedTuple):
    size: int
    modification_time: float
    settling_time: Optional[float] = None


class DirWatcher(murfey.util.Observer):
    def __init__(
        self,
        path: str | os.PathLike,
        settling_time: float = 60,
        status_bar: StatusBar | None = None,
    ):
        super().__init__()
        self._basepath = os.fspath(path)
        self._lastscan: dict[str, _FileInfo] | None = {}
        self._file_candidates: dict[str, _FileInfo] = {}
        self._statusbar = status_bar
        self.settling_time = settling_time
        self._modification_overwrite: float | None = None
        self._init_time: float = time.time()

    def __repr__(self) -> str:
        return f"<DirWatcher ({self._basepath})>"

    def scan(self, modification_time: float | None = None, transfer_all: bool = False):
        try:
            filelist = self._scan_directory(
                modification_time=self._modification_overwrite or modification_time
            )
            scan_completion = time.time()

            for entry, entry_info in filelist.items():
                if self._lastscan is not None and entry_info != self._lastscan.get(
                    entry
                ):
                    self._file_candidates[entry] = entry_info._replace(
                        settling_time=scan_completion
                    )

            for x in sorted(
                self._file_candidates,
                key=lambda _x: self._file_candidates[_x].modification_time,
            ):
                if x not in filelist:
                    log.info(f"Previously seen file {x!r} has disappeared")
                    del self._file_candidates[x]
                    continue

                if (
                    self._file_candidates[x].settling_time + self.settling_time  # type: ignore
                    < time.time()
                ):
                    try:
                        file_stat = os.stat(x)
                        if (
                            file_stat.st_size == self._file_candidates[x].size
                            and file_stat.st_mtime
                            <= self._file_candidates[x].modification_time
                            and file_stat.st_ctime
                            <= self._file_candidates[x].modification_time
                        ):
                            if (
                                not modification_time
                                and not self._modification_overwrite
                                and not transfer_all
                            ):
                                if file_stat.st_mtime >= self._init_time:
                                    top_level_dir = (
                                        Path(self._basepath)
                                        / Path(x).relative_to(self._basepath).parts[0]
                                    )
                                    if top_level_dir.is_dir():
                                        # touch the changing directory so that when _modification_overwrite is set
                                        # we don't potentially catch old directories that aren't changing
                                        # this means it will only autodetect new directories from this point
                                        top_level_dir.touch(exist_ok=True)
                                        filelist.update(
                                            self._scan_directory(
                                                path=str(top_level_dir)
                                            )
                                        )
                                        self._modification_overwrite = max(
                                            top_level_dir.stat().st_mtime,
                                            top_level_dir.stat().st_ctime,
                                        )
                            else:
                                self._notify_for_transfer(x)
                            continue
                    except Exception as e:
                        log.error(f"Exception encountered: {e}", exc_info=True)
                        return

                if self._lastscan is not None and x not in self._lastscan:
                    log.debug(
                        f"Found file {Path(x).name!r} for potential future transfer"
                    )

            self._lastscan = filelist
        except Exception as e:
            log.error(f"Exception encountered: {e}")

    def _notify_for_transfer(self, file_candidate: str):
        log.debug(f"File {Path(file_candidate).name!r} is ready to be transferred")
        if self._statusbar:
            # log.info("Increasing number to be transferred")
            with self._statusbar.lock:
                self._statusbar.transferred = [
                    self._statusbar.transferred[0],
                    self._statusbar.transferred[1] + 1,
                ]

        self.notify(Path(file_candidate))
        del self._file_candidates[file_candidate]

    def _scan_directory(
        self, path: str = "", modification_time: float | None = None
    ) -> dict[str, _FileInfo]:
        result: dict[str, _FileInfo] = {}
        try:
            directory_contents = os.scandir(os.path.join(self._basepath, path))
        except FileNotFoundError:
            # Possible race condition here if the directory disappears before
            # we had a chance to scan it. If it is a sub directory then we can
            # just ignore this case, but if it is our main directory we should
            # raise it to the caller.
            if path:
                return result
            raise
        for entry in directory_contents:
            entry_name = os.path.join(path, entry.name)
            if entry.is_dir() and (
                modification_time is None or entry.stat().st_ctime >= modification_time
            ):
                result.update(self._scan_directory(entry_name))
            else:
                # avoid textual log
                if "textual" in str(entry):
                    continue
                try:
                    file_stat = entry.stat()
                except FileNotFoundError:
                    # Possible race condition here if the file disappears
                    # between the scandir and the stat call.
                    # In this case we can just ignore the file.
                    continue
                if modification_time:
                    if max(file_stat.st_mtime, file_stat.st_ctime) >= modification_time:
                        result[str(Path(self._basepath) / entry_name)] = _FileInfo(
                            size=file_stat.st_size,
                            modification_time=max(
                                file_stat.st_mtime, file_stat.st_ctime
                            ),
                        )
                else:
                    result[str(Path(self._basepath) / entry_name)] = _FileInfo(
                        size=file_stat.st_size,
                        modification_time=max(file_stat.st_mtime, file_stat.st_ctime),
                    )
        return result



from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Optional, Union

import yaml
from pydantic import BaseModel


class MachineConfig(BaseModel):
    acquisition_software: List[str]
    calibrations: Dict[str, Union[dict, float]]
    data_directories: Dict[Path, str]
    rsync_basepath: Path
    rsync_module: str = ""
    gain_reference_directory: Optional[Path] = None
    processed_directory_name: str = "processed"
    feedback_queue: str = "murfey_feedback"


def from_file(config_file_path: Path, microscope: str) -> MachineConfig:
    with open(config_file_path, "r") as config_stream:
        config = yaml.safe_load(config_stream)
    return MachineConfig(**config.get(microscope, {}))



from __future__ import annotations

import argparse
import logging
import os
import socket
from functools import lru_cache, singledispatch
from pathlib import Path
from threading import Thread
from typing import Any

import uvicorn
import workflows
import zocalo.configuration
from fastapi.templating import Jinja2Templates
from ispyb.sqlalchemy._auto_db_schema import (
    AutoProcProgram,
    Base,
    DataCollection,
    DataCollectionGroup,
    ProcessingJob,
)
from rich.logging import RichHandler
from sqlalchemy.exc import SQLAlchemyError

import murfey
import murfey.server.websocket
from murfey.server.config import MachineConfig, from_file

try:
    from murfey.server.ispyb import TransportManager  # Session
except AttributeError:
    pass
from murfey.util.state import global_state

try:
    from importlib.resources import files  # type: ignore
except ImportError:
    # Fallback for Python 3.8
    from importlib_resources import files  # type: ignore

logger = logging.getLogger("murfey.server")

template_files = files("murfey") / "templates"
templates = Jinja2Templates(directory=template_files)

_running_server: uvicorn.Server | None = None
_transport_object = None


def respond_with_template(filename: str, parameters: dict[str, Any] | None = None):
    template_parameters = {
        "hostname": get_hostname(),
        "microscope": get_microscope(),
        "version": murfey.__version__,
    }
    if parameters:
        template_parameters.update(parameters)
    return templates.TemplateResponse(filename, template_parameters)


class LogFilter(logging.Filter):
    """A filter to limit messages going to Graylog"""

    def __repr__(self):
        return "<murfey.server.LogFilter>"

    def __init__(self):
        self._filter_levels = {
            "murfey": logging.DEBUG,
            "ispyb": logging.DEBUG,
            "zocalo": logging.DEBUG,
            "uvicorn": logging.INFO,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.INFO,
        }

    @staticmethod
    def install() -> LogFilter:
        logfilter = LogFilter()
        root_logger = logging.getLogger()
        for handler in root_logger.handlers:
            handler.addFilter(logfilter)
        return logfilter

    def filter(self, record: logging.LogRecord) -> bool:
        logger_name = record.name
        while True:
            if logger_name in self._filter_levels:
                return record.levelno >= self._filter_levels[logger_name]
            if "." not in logger_name:
                return False
            logger_name = logger_name.rsplit(".", maxsplit=1)[0]


def run():
    parser = argparse.ArgumentParser(description="Start the Murfey server")
    parser.add_argument(
        "--host",
        help="Listen for incoming connections on a specific interface (IP address or hostname; default: all)",
        default="0.0.0.0",
    )
    parser.add_argument(
        "--port",
        help="Listen for incoming TCP connections on this port (default: 8000)",
        type=int,
        default=8000,
    )
    parser.add_argument(
        "--demo",
        action="store_true",
    )
    parser.add_argument(
        "--feedback",
        action="store_true",
    )

    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=False,
        help="Decrease logging output verbosity",
    )
    verbosity.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Increase logging output verbosity",
        default=0,
    )

    # setup logging
    zc = zocalo.configuration.from_file()
    zc.activate()

    # Install a log filter to all existing handlers.
    # At this stage this will exclude console loggers, but will cover
    # any Graylog logging set up by the environment activation
    LogFilter.install()

    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)

    args = parser.parse_args()

    # Set up Zocalo connection
    if args.demo:
        os.environ["MURFEY_DEMO"] = "1"
    else:
        _set_up_transport(args.transport)

    # Set up logging now that the desired verbosity is known
    _set_up_logging(quiet=args.quiet, verbosity=args.verbose)

    murfey_machine_configuration = os.environ["MURFEY_MACHINE_CONFIGURATION"]
    machine_config: MachineConfig = MachineConfig(
        acquisition_software=[],
        calibrations={},
        data_directories={},
        rsync_basepath=Path("dls/tmp"),
    )
    if murfey_machine_configuration:
        microscope = get_microscope()
        machine_config = from_file(Path(murfey_machine_configuration), microscope)

    rabbit_thread = Thread(
        target=feedback_listen,
        kwargs={"feedback_queue": machine_config.feedback_queue},
        daemon=True,
    )
    logger.info("Starting Murfey RabbitMQ thread")
    if args.feedback:
        rabbit_thread.start()

    logger.info(
        f"Starting Murfey server version {murfey.__version__} for beamline {get_microscope()}, listening on {args.host}:{args.port}"
    )
    global _running_server
    config = uvicorn.Config(
        "murfey.server.main:app",
        host=args.host,
        port=args.port,
        log_config=None,
    )

    _running_server = uvicorn.Server(config=config)
    _running_server.run()
    logger.info("Server shutting down")


def shutdown():
    global _running_server
    if _running_server:
        _running_server.should_exit = True
        _running_server.force_exit = True


@lru_cache()
def get_microscope():
    try:
        hostname = get_hostname()
        microscope_from_hostname = hostname.split(".")[0]
    except OSError:
        microscope_from_hostname = "Unknown"
    microscope_name = os.getenv("BEAMLINE", microscope_from_hostname)
    return microscope_name


@lru_cache()
def get_hostname():
    return socket.gethostname()


def _set_up_logging(quiet: bool, verbosity: int):
    rich_handler = RichHandler(enable_link_path=False)
    if quiet:
        rich_handler.setLevel(logging.INFO)
        log_levels = {
            "murfey": logging.INFO,
            "uvicorn": logging.WARNING,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.WARNING,
        }
    elif verbosity <= 0:
        rich_handler.setLevel(logging.INFO)
        log_levels = {
            "murfey": logging.DEBUG,
            "uvicorn": logging.INFO,
            "uvicorn.access": logging.WARNING,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.WARNING,
        }
    elif verbosity <= 1:
        rich_handler.setLevel(logging.DEBUG)
        log_levels = {
            "": logging.INFO,
            "murfey": logging.DEBUG,
            "uvicorn": logging.INFO,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.WARNING,
        }
    elif verbosity <= 2:
        rich_handler.setLevel(logging.DEBUG)
        log_levels = {
            "": logging.INFO,
            "murfey": logging.DEBUG,
            "uvicorn": logging.DEBUG,
            "fastapi": logging.DEBUG,
            "starlette": logging.DEBUG,
            "sqlalchemy": logging.WARNING,
        }
    else:
        rich_handler.setLevel(logging.DEBUG)
        log_levels = {
            "": logging.DEBUG,
            "murfey": logging.DEBUG,
            "uvicorn": logging.DEBUG,
            "fastapi": logging.DEBUG,
            "starlette": logging.DEBUG,
            "sqlalchemy": logging.DEBUG,
        }

    logging.getLogger().addHandler(rich_handler)
    for logger_name, log_level in log_levels.items():
        logging.getLogger(logger_name).setLevel(log_level)


def _set_up_transport(transport_type):
    global _transport_object
    _transport_object = TransportManager(transport_type)


async def feedback_callback_async(header: dict, message: dict) -> None:
    logger.info(f"feedback_callback_async called with {header}, {message}")
    if message["register"] == "motion_corrected":
        if murfey.server.websocket.manager:
            if global_state.get("motion_corrected_movies") and isinstance(
                global_state["motion_corrected_movies"], dict
            ):
                await global_state.aupdate(
                    "motion_corrected_movies",
                    {
                        message.get("movie"): [
                            message.get("mrc_out"),
                            message.get("movie_id"),
                        ],
                    },
                )
            else:
                await global_state.aupdate(
                    "motion_corrected_movies",
                    {
                        message.get("movie"): [
                            message.get("mrc_out"),
                            message.get("movie_id"),
                        ]
                    },
                )


def feedback_callback(header: dict, message: dict) -> None:
    record = None
    if "environment" in message:
        message = message["payload"]
    if message["register"] == "motion_corrected":
        global_state.update(
            "motion_corrected_movies",
            {
                message.get("movie"): [
                    message.get("mrc_out"),
                    message.get("movie_id"),
                ]
            },
            perform_state_update=False,
        )

        if _transport_object:
            _transport_object.transport.ack(header)
        return None
    elif message["register"] == "data_collection_group":
        record = DataCollectionGroup(
            sessionId=message["session_id"],
            experimentType=message["experiment_type"],
            experimentTypeId=message["experiment_type_id"],
        )
        print(record)
        dcgid = _register(record, header)
        print(dcgid)
        if _transport_object:
            if dcgid is None:
                _transport_object.transport.nack(header)
                return None
            global_state["data_collection_group_id"] = dcgid
        _transport_object.transport.ack(header)
        return None
    elif message["register"] == "data_collection":
        record = DataCollection(
            SESSIONID=message["session_id"],
            experimenttype=message["experiment_type"],
            imageDirectory=message["image_directory"],
            imageSuffix=message["image_suffix"],
            voltage=message["voltage"],
            dataCollectionGroupId=global_state.get("data_collection_group_id"),
        )
        dcid = _register(record, header)
        if dcid is None and _transport_object:
            _transport_object.transport.nack(header)
            return None
        logger.debug(f"registered: {message.get('tag')}")
        if global_state.get("data_collection_ids") and isinstance(
            global_state["data_collection_ids"], dict
        ):
            global_state["data_collection_ids"] = {
                **global_state["data_collection_ids"],
                message.get("tag"): dcid,
            }
        else:
            global_state["data_collection_ids"] = {message.get("tag"): dcid}
        if _transport_object:
            _transport_object.transport.ack(header)
        return None
    elif message["register"] == "processing_job":
        assert isinstance(global_state["data_collection_ids"], dict)
        _dcid = global_state["data_collection_ids"][message["tag"]]
        record = ProcessingJob(dataCollectionId=_dcid, recipe=message["recipe"])
        pid = _register(record, header)
        if pid is None and _transport_object:
            _transport_object.transport.nack(header)
            return None
        if global_state.get("processing_job_ids"):
            global_state["processing_job_ids"] = {
                **global_state["processing_job_ids"],  # type: ignore
                message.get("tag"): {
                    **global_state["processing_job_ids"].get(message.get("tag"), {}),  # type: ignore
                    message["recipe"]: pid,
                },
            }
        else:
            prids = {message["tag"]: {message["recipe"]: pid}}
            global_state["processing_job_ids"] = prids
        record = AutoProcProgram(processingJobId=pid)
        appid = _register(record, header)
        if appid is None and _transport_object:
            _transport_object.transport.nack(header)
            return None
        if global_state.get("autoproc_program_ids"):
            assert isinstance(global_state["autoproc_program_ids"], dict)
            global_state["autoproc_program_ids"] = {
                **global_state["autoproc_program_ids"],
                message.get("tag"): {
                    **global_state["processing_job_ids"].get(message.get("tag"), {}),  # type: ignore
                    message["recipe"]: appid,
                },
            }
        else:
            global_state["autoproc_program_ids"] = {
                message["tag"]: {message["recipe"]: appid}
            }
        if _transport_object:
            _transport_object.transport.ack(header)
        return None
    if _transport_object:
        _transport_object.transport.nack(header, requeue=False)
    return None


@singledispatch
def _register(record, header: dict):
    raise NotImplementedError(f"Not method to register {record} or type {type(record)}")


@_register.register
def _(record: Base, header: dict):
    if not _transport_object:
        logger.error(
            f"No transport object found when processing record {record}. Message header: {header}"
        )
        return None
    try:
        if isinstance(record, DataCollection):
            return _transport_object.do_insert_data_collection(record)["return_value"]
        if isinstance(record, DataCollectionGroup):
            return _transport_object.do_insert_data_collection_group(record)[
                "return_value"
            ]
        if isinstance(record, ProcessingJob):
            return _transport_object.do_create_ispyb_job(record)["return_value"]
        if isinstance(record, AutoProcProgram):
            return _transport_object.do_update_processing_status(record)["return_value"]
        # session = Session()
        # session.add(record)
        # session.commit()
        # _transport_object.transport.ack(header, requeue=False)
        return getattr(record, record.__table__.primary_key.columns[0].name)

    except SQLAlchemyError as e:
        logger.error(f"Murfey failed to insert ISPyB record {record}", e, exc_info=True)
        # _transport_object.transport.nack(header)
        return None
    except AttributeError as e:
        logger.error(
            f"Murfey could not find primary key when inserting record {record}",
            e,
            exc_info=True,
        )
        return None


def feedback_listen(feedback_queue: str = "murfey_feedback"):
    if _transport_object:
        _transport_object.transport.subscribe(
            feedback_queue, feedback_callback, acknowledgement=True
        )



from __future__ import annotations

import argparse
import configparser

# import json
import logging
import os
import platform
import shutil
import sys
import time
import webbrowser
from datetime import datetime
from pathlib import Path
from queue import Queue

# from multiprocessing import Process, Queue
from threading import Thread
from typing import Literal
from urllib.parse import ParseResult, urlparse

import procrunner
import requests

import murfey.client.rsync
import murfey.client.update
import murfey.client.watchdir
import murfey.client.websocket
from murfey.client.analyser import Analyser
from murfey.client.customlogging import CustomHandler, DirectableRichHandler
from murfey.client.gain_ref import determine_gain_ref
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.client.rsync import RSyncer
from murfey.client.tui.app import MurfeyTUI
from murfey.client.tui.status_bar import StatusBar
from murfey.util.models import Visit

# from asyncio import Queue


# from rich.prompt import Prompt


log = logging.getLogger("murfey.client")


def _enable_webbrowser_in_cygwin():
    """Helper function to make webbrowser.open() work in CygWin"""
    if "cygwin" in platform.system().lower() and shutil.which("cygstart"):
        webbrowser.register("cygstart", None, webbrowser.GenericBrowser("cygstart"))


def _check_for_updates(
    server: ParseResult, install_version: None | Literal[True] | str
):
    if install_version is True:
        # User requested installation of the newest version
        try:
            murfey.client.update.check(server, force=True)
            print("\nYou are already running the newest version of Murfey")
            exit()
        except Exception as e:
            exit(f"Murfey update check failed with {e}")

    if install_version:
        # User requested installation of a specific version
        if murfey.client.update.install_murfey(server, install_version):
            print(f"\nMurfey has been updated to version {install_version}")
            exit()
        else:
            exit("Error occurred while updating Murfey")

    # Otherwise run a routine update check to ensure client and server are compatible
    try:
        murfey.client.update.check(server)
    except Exception as e:
        print(f"Murfey update check failed with {e}")


def _get_visit_list(api_base: ParseResult, demo: bool = False):
    get_visits_url = api_base._replace(path="/visits_raw")
    server_reply = requests.get(get_visits_url.geturl())
    if server_reply.status_code != 200:
        raise ValueError(f"Server unreachable ({server_reply.status_code})")
    return [Visit.parse_obj(v) for v in server_reply.json()]


def run():
    config = read_config()
    try:
        server_routing = config["ServerRouter"]
    except KeyError:
        server_routing = {}
    server_routing_prefix_found = False
    if server_routing:
        for path_prefix, server in server_routing.items():
            if str(Path.cwd()).startswith(path_prefix):
                known_server = server
                server_routing_prefix_found = True
                break
            else:
                known_server = None
    else:
        known_server = config["Murfey"].get("server")

    parser = argparse.ArgumentParser(description="Start the Murfey client")
    parser.add_argument(
        "--server",
        metavar="HOST:PORT",
        type=str,
        help=f"Murfey server to connect to ({known_server})",
        default=known_server,
    )
    parser.add_argument("--visit", help="Name of visit")
    parser.add_argument(
        "--source", help="Directory to transfer files from", type=Path, default="."
    )
    parser.add_argument(
        "--destination",
        help="Directory to transfer files to (syntax: 'data/2022/cm31093-2/tmp/murfey')",
    )
    parser.add_argument(
        "--update",
        metavar="VERSION",
        nargs="?",
        default=None,
        const=True,
        help="Update Murfey to the newest or to a specific version",
    )
    parser.add_argument(
        "--demo",
        action="store_true",
    )
    parser.add_argument(
        "--appearance_time",
        type=float,
        default=-1,
        help="Only consider top level directories that have appeared more recently than this many hours ago",
    )
    parser.add_argument(
        "--real_dc",
        action="store_true",
        default=False,
        help="Actually perform data collection related calls to API (will do inserts in ISPyB)",
    )
    parser.add_argument(
        "--transfer_all",
        action="store_true",
        help="Transfer all files in current directory regardless of age",
    )
    parser.add_argument(
        "--no_transfer",
        action="store_true",
        help="Avoid actually transferring files",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Turn on debugging logs",
    )
    parser.add_argument(
        "--local",
        action="store_true",
        default=False,
        help="Perform rsync transfers locally rather than remotely",
    )
    parser.add_argument(
        "--force_mdoc_metadata",
        action="store_true",
        default=False,
        help="Force metadata to be read from mdoc files",
    )

    args = parser.parse_args()

    if not args.server:
        exit("Murfey server not set. Please run with --server host:port")
    if not args.server.startswith(("http://", "https://")):
        if "://" in args.server:
            exit("Unknown server protocol. Only http:// and https:// are allowed")
        args.server = f"http://{args.server}"

    murfey_url = urlparse(args.server, allow_fragments=False)
    if args.server != known_server:
        # New server specified. Verify that it is real
        print(f"Attempting to connect to new server {args.server}")
        try:
            murfey.client.update.check(murfey_url, install=False)
        except Exception as e:
            exit(f"Could not reach Murfey server at {args.server!r} - {e}")

        # If server is reachable then update the configuration
        config["Murfey"]["server"] = args.server
        write_config(config)

    # If user requested installation of a specific or a newer version then
    # make that happen, otherwise ensure client and server are compatible and
    # update if necessary.
    _check_for_updates(server=murfey_url, install_version=args.update)

    if args.no_transfer:
        log.info("No files will be transferred as --no_transfer flag was specified")

    from pprint import pprint

    ongoing_visits = []
    if args.visit:
        ongoing_visits = [args.visit]
    elif server_routing_prefix_found:
        for part in Path.cwd().parts:
            if "-" in part:
                ongoing_visits = [part]
                break
    if not ongoing_visits:
        print("Ongoing visits:")
        ongoing_visits = _get_visit_list(murfey_url, demo=args.demo)
        pprint(ongoing_visits)
        ongoing_visits = [v.name for v in ongoing_visits]

    _enable_webbrowser_in_cygwin()

    log.setLevel(logging.DEBUG)
    log_queue = Queue()
    input_queue = Queue()

    rich_handler = DirectableRichHandler(log_queue, enable_link_path=False)
    rich_handler.setLevel(logging.DEBUG if args.debug else logging.INFO)
    ws = murfey.client.websocket.WSApp(server=args.server)
    logging.getLogger().addHandler(rich_handler)
    handler = CustomHandler(ws.send)
    logging.getLogger().addHandler(handler)
    logging.getLogger("murfey").setLevel(logging.INFO)
    logging.getLogger("websocket").setLevel(logging.WARNING)

    log.info("Starting Websocket connection")

    # start_dc = Prompt.ask("Would you like to register a new data collection?", choices=["y", "n"])

    # if start_dc == "y":
    #     image_directory = str(args.destination)
    #     image_suffix = Prompt.ask("Enter the image suffix", choices=[".tiff", ".tif", ".mrc", ".eer"])
    #     visit = str(args.visit)
    #     dc_params = {
    #         "type": "start_dc",
    #         "image_directory": image_directory,
    #         "image_suffix": image_suffix,
    #         "visit": visit,
    #     }
    #     ws.send(json.dumps(dc_params))

    status_bar = StatusBar()
    source_watcher = murfey.client.watchdir.DirWatcher(
        args.source, settling_time=1, status_bar=status_bar
    )

    machine_data = requests.get(f"{murfey_url.geturl()}/machine/").json()
    gain_ref: Path | None = None
    if machine_data.get("gain_reference_directory"):
        try:
            gain_ref = determine_gain_ref(machine_data["gain_reference_directory"])
        except RuntimeError:
            pass

    main_loop_thread = Thread(
        target=main_loop,
        args=[source_watcher, args.appearance_time, args.transfer_all],
        kwargs={"gain_ref": gain_ref},
        daemon=True,
    )
    main_loop_thread.start()

    instance_environment = MurfeyInstanceEnvironment(
        url=murfey_url,
        source=Path(args.source),
        watcher=source_watcher,
        default_destination=args.destination
        or f"{machine_data.get('rsync_module') or 'data'}/{datetime.now().year}",
        demo=args.demo,
        processing_only_mode=server_routing_prefix_found,
    )

    ws.environment = instance_environment

    rsync_process = RSyncer(
        instance_environment.source,
        basepath_remote=Path(args.destination or f"data/{datetime.now().year}"),
        server_url=murfey_url,
        local=args.local or instance_environment.demo,
        do_transfer=not args.no_transfer,
    )
    source_watcher.subscribe(rsync_process.enqueue)

    analyser = Analyser(
        instance_environment.source,
        environment=instance_environment if args.real_dc else None,
        force_mdoc_metadata=args.force_mdoc_metadata,
    )
    # source_watcher.subscribe(analyser.enqueue)
    rsync_process.subscribe(analyser.enqueue)

    rich_handler.redirect = True
    MurfeyTUI.run(
        log_verbosity=2,
        environment=instance_environment,
        visits=ongoing_visits,
        queues={"input": input_queue, "logs": log_queue},
        status_bar=status_bar,
        dummy_dc=not args.real_dc,
        do_transfer=not args.no_transfer,
        rsync_process=rsync_process,
        analyser=analyser,
    )
    rich_handler.redirect = False

    try:
        main_loop_thread.join()
    except KeyboardInterrupt:
        log.info("Encountered CTRL+C")
        # if args.destination:
        #     rsync_process.stop()
        ws.close()
        log.info("Client stopped")


def main_loop(
    source_watcher: murfey.client.watchdir.DirWatcher,
    appearance_time: float,
    transfer_all: bool,
    gain_ref: Path | None = None,
):
    log.info(
        f"Murfey {murfey.__version__} on Python {'.'.join(map(str, sys.version_info[0:3]))} entering main loop"
    )
    if gain_ref:
        gain_rsync = procrunner.run(["rsync", str(gain_ref)])
        if gain_rsync.returncode:
            log.warning(
                f"Gain reference file {gain_ref} was not successfully transferred"
            )
    if appearance_time > 0:
        modification_time: float | None = time.time() - appearance_time * 3600
    else:
        modification_time = None
    while True:
        source_watcher.scan(
            modification_time=modification_time, transfer_all=transfer_all
        )
        time.sleep(15)


def read_config() -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    try:
        mcch = os.environ.get("MURFEY_CLIENT_CONFIG_HOME")
        murfey_client_config_home = Path(mcch) if mcch else Path.home()
        with open(murfey_client_config_home / ".murfey") as configfile:
            config.read_file(configfile)
    except FileNotFoundError:
        pass
    if "Murfey" not in config:
        config["Murfey"] = {}
    return config


def write_config(config: configparser.ConfigParser):
    mcch = os.environ.get("MURFEY_CLIENT_CONFIG_HOME")
    murfey_client_config_home = Path(mcch) if mcch else Path.home()
    with open(murfey_client_config_home / ".murfey", "w") as configfile:
        config.write(configfile)



from __future__ import annotations

import logging
from pathlib import Path
from threading import RLock
from typing import Callable, Dict, List, OrderedDict

import requests
import xmltodict
from pydantic import BaseModel

from murfey.client.instance_environment import (
    MovieID,
    MovieTracker,
    MurfeyID,
    MurfeyInstanceEnvironment,
    global_env_lock,
)
from murfey.client.tui.forms import TUIFormValue
from murfey.util.mdoc import get_global_data

# import time

logger = logging.getLogger("murfey.client.context")


def detect_acquisition_software(dir_for_transfer: Path) -> str:
    glob = dir_for_transfer.glob("*")
    for f in glob:
        if f.name.startswith("EPU") or f.name.startswith("GridSquare"):
            return "epu"
        if f.name.startswith("Position") or f.suffix == ".mdoc":
            return "tomo"
    return ""


class Context:
    def __init__(self, acquisition_software: str):
        self._acquisition_software = acquisition_software

    def post_transfer(self, transferred_file: Path, role: str = "", **kwargs):
        raise NotImplementedError(
            f"post_transfer hook must be declared in derived class to be used: {self}"
        )

    def post_first_transfer(self, transferred_file: Path, role: str = "", **kwargs):
        self.post_transfer(transferred_file, role=role, **kwargs)

    def gather_metadata(self, metadata_file: Path):
        raise NotImplementedError(
            f"gather_metadata must be declared in derived class to be used: {self}"
        )


class SPAContext(Context):
    def post_transfer(self, transferred_file: Path, role: str = "", **kwargs):
        pass


class ProcessFileIncomplete(BaseModel):
    dest: Path
    source: Path
    image_number: int
    mc_uuid: int
    tag: str
    description: str = ""


class TomographyContext(Context):
    def __init__(self, acquisition_software: str):
        super().__init__(acquisition_software)
        self._tilt_series: Dict[str, List[Path]] = {}
        self._completed_tilt_series: List[str] = []
        self._aligned_tilt_series: List[str] = []
        self._motion_corrected_tilt_series: Dict[str, List[Path]] = {}
        self._last_transferred_file: Path | None = None
        self._data_collection_stash: list = []
        self._processing_job_stash: dict = {}
        self._preprocessing_triggers: dict = {}
        self._lock: RLock = RLock()
        self._extract_tilt_series: Callable[[Path], str] | None = None
        self._extract_tilt_tag: Callable[[Path], str] | None = None

    def _flush_data_collections(self):
        logger.info("Flushing data collection API calls")
        for dc_data in self._data_collection_stash:
            data = {**dc_data[2], **dc_data[1].data_collection_parameters}
            requests.post(dc_data[0], json=data)
        self._data_collection_stash = []

    def _flush_processing_job(self, tag: str):
        # logger.info(
        #     f"Flushing processing job {tag}, {self._processing_job_stash.get(tag)}"
        # )
        if proc_data := self._processing_job_stash.get(tag):
            for pd in proc_data:
                requests.post(pd[0], json=pd[1])
            self._processing_job_stash.pop(tag)

    def _flush_preprocess(self, tag: str, app_id: int):
        # logger.info(f"Flushing preprocessing requests {tag}")
        if tag_tr := self._preprocessing_triggers.get(tag):
            for tr in tag_tr:
                process_file = self._complete_process_file(tr[1], tr[2], app_id)
                if process_file:
                    requests.post(tr[0], json=process_file)
            self._preprocessing_triggers.pop(tag)

    def _check_for_alignment(
        self,
        movie_path: Path,
        motion_corrected_path: Path,
        url: str,
        dcid: int,
        pjid: int,
        appid: int,
        mvid: int,
        tilt_angles: List,
    ):
        if self._extract_tilt_series and self._extract_tilt_tag:
            tilt_series = self._extract_tilt_tag(
                movie_path
            ) + self._extract_tilt_series(movie_path)
        else:
            return

        if self._motion_corrected_tilt_series.get(
            tilt_series
        ) and motion_corrected_path not in self._motion_corrected_tilt_series.get(
            tilt_series, {}
        ):
            self._motion_corrected_tilt_series[tilt_series].append(
                motion_corrected_path
            )
        else:
            self._motion_corrected_tilt_series[tilt_series] = [motion_corrected_path]
        if tilt_series in self._completed_tilt_series:
            if (
                len(self._motion_corrected_tilt_series[tilt_series])
                == len(self._tilt_series[tilt_series])
                and len(self._motion_corrected_tilt_series[tilt_series]) > 1
                and tilt_series not in self._aligned_tilt_series
            ):
                try:

                    series_data: dict = {
                        "name": tilt_series,
                        "file_tilt_list": str(tilt_angles),
                        "dcid": dcid,
                        "processing_job": pjid,
                        "autoproc_program_id": appid,
                        "motion_corrected_path": str(motion_corrected_path),
                        "movie_id": mvid,
                    }
                    requests.post(url, json=series_data)
                    with self._lock:
                        self._aligned_tilt_series.append(tilt_series)
                except Exception as e:
                    logger.warning(f"Data error {e}")

    def _complete_process_file(
        self,
        incomplete_process_file: ProcessFileIncomplete,
        environment: MurfeyInstanceEnvironment,
        app_id: int,
    ) -> dict:
        try:
            with global_env_lock:
                tag = incomplete_process_file.tag

                new_dict = {
                    "path": str(incomplete_process_file.dest),
                    "description": incomplete_process_file.description,
                    "size": incomplete_process_file.source.stat().st_size,
                    "timestamp": incomplete_process_file.source.stat().st_ctime,
                    "processing_job": environment.processing_job_ids[tag][
                        "em-tomo-preprocess"
                    ],
                    "data_collection_id": environment.data_collection_ids[tag],
                    "image_number": incomplete_process_file.image_number,
                    "pixel_size": environment.data_collection_parameters[
                        "pixel_size_on_image"
                    ],
                    "autoproc_program_id": app_id,
                    "mc_uuid": incomplete_process_file.mc_uuid,
                    "gain_ref": environment.data_collection_parameters.get("gain_ref"),
                }
                return new_dict
        except KeyError:
            logger.warning("Key error encountered in _complete_process_file")
            return {}

    def _add_tilt(
        self,
        file_path: Path,
        extract_tilt_series: Callable[[Path], str],
        extract_tilt_angle: Callable[[Path], str],
        extract_tilt_tag: Callable[[Path], str],
        environment: MurfeyInstanceEnvironment | None = None,
    ) -> List[str]:
        if not self._extract_tilt_series:
            self._extract_tilt_series = extract_tilt_series
        if not self._extract_tilt_tag:
            self._extract_tilt_tag = extract_tilt_tag
        # time.sleep(5)
        try:
            tilt_series_num = extract_tilt_series(file_path)
            tilt_angle = extract_tilt_angle(file_path)
            tilt_tag = extract_tilt_tag(file_path)
            try:
                float(tilt_series_num)
                float(tilt_angle)
            except ValueError:
                return []
            tilt_series = tilt_tag + tilt_series_num

        except Exception:
            logger.info(
                f"Tilt series and angle could not be determined for {file_path}"
            )
            return []

        if environment:
            machine_config = (
                {}
                if environment.demo
                else requests.get(f"{str(environment.url.geturl())}/machine/").json()
            )
            if environment.visit in environment.default_destination:
                file_transferred_to = (
                    Path(machine_config.get("rsync_basepath", ""))
                    / Path(environment.default_destination)
                    / file_path.name
                )
            else:
                file_transferred_to = (
                    Path(machine_config.get("rsync_basepath", ""))
                    / Path(environment.default_destination)
                    / environment.visit
                    / file_path.name
                )
            environment.movies[file_transferred_to] = MovieTracker(
                movie_number=next(MovieID),
                motion_correction_uuid=next(MurfeyID),
            )
            environment.movie_tilt_pair[file_transferred_to] = tilt_series
            if environment.tilt_angles.get(tilt_series):
                environment.tilt_angles[tilt_series].append(
                    [str(file_transferred_to), tilt_angle]
                )
            else:
                environment.tilt_angles[tilt_series] = [
                    [str(file_transferred_to), tilt_angle]
                ]
        if tilt_series in self._completed_tilt_series:
            logger.info(
                f"Tilt series {tilt_series} was previously thought complete but now {file_path} has been seen"
            )
            self._completed_tilt_series.remove(tilt_series)
            if tilt_series in self._aligned_tilt_series:
                with self._lock:
                    self._aligned_tilt_series.remove(tilt_series)

        if not self._tilt_series.get(tilt_series):
            logger.info(f"New tilt series found: {tilt_series}")
            self._tilt_series[tilt_series] = [file_path]
            try:
                if environment:
                    url = f"{str(environment.url.geturl())}/visits/{environment.visit}/start_data_collection"
                    data = {
                        "experiment_type": "tomography",
                        "tilt": tilt_series_num,
                        "file_extension": file_path.suffix,
                        "acquisition_software": self._acquisition_software,
                        "image_directory": str(file_path.parent),
                        "tag": tilt_series,
                    }
                    if environment.data_collection_parameters:
                        data.update(
                            {
                                "voltage": environment.data_collection_parameters[
                                    "voltage"
                                ],
                                "pixel_size_on_image": environment.data_collection_parameters[
                                    "pixel_size_on_image"
                                ],
                                "image_size_x": environment.data_collection_parameters[
                                    "image_size_x"
                                ],
                                "image_size_y": environment.data_collection_parameters[
                                    "image_size_y"
                                ],
                            }
                        )
                    if environment.data_collection_group_id is None:
                        self._data_collection_stash.append((url, environment, data))
                    else:
                        requests.post(url, json=data)
                    proc_url = f"{str(environment.url.geturl())}/visits/{environment.visit}/register_processing_job"
                    if environment.data_collection_ids.get(tilt_series) is None:
                        self._processing_job_stash[tilt_series] = [
                            (
                                proc_url,
                                {"tag": tilt_series, "recipe": "em-tomo-preprocess"},
                            )
                        ]
                        self._processing_job_stash[tilt_series].append(
                            (proc_url, {"tag": tilt_series, "recipe": "em-tomo-align"})
                        )
                    else:
                        if self._processing_job_stash.get(tilt_series):
                            self._flush_processing_job(tilt_series)
                        requests.post(
                            proc_url,
                            json={"tag": tilt_series, "recipe": "em-tomo-preprocess"},
                        )
                        requests.post(
                            proc_url,
                            json={"tag": tilt_series, "recipe": "em-tomo-align"},
                        )
            except Exception as e:
                logger.error(f"ERROR {e}")
        else:
            if file_path not in self._tilt_series[tilt_series]:
                self._tilt_series[tilt_series].append(file_path)

        if environment and environment.autoproc_program_ids.get(tilt_series):
            preproc_url = f"{str(environment.url.geturl())}/visits/{environment.visit}/tomography_preprocess"
            preproc_data = {
                "path": str(file_transferred_to),
                "description": "",
                "size": file_path.stat().st_size,
                "timestamp": file_path.stat().st_ctime,
                "processing_job": environment.processing_job_ids[tilt_series][
                    "em-tomo-preprocess"
                ],
                "data_collection_id": environment.data_collection_ids[tilt_series],
                "image_number": environment.movies[file_transferred_to].movie_number,
                "pixel_size": environment.data_collection_parameters[
                    "pixel_size_on_image"
                ],
                "autoproc_program_id": environment.autoproc_program_ids[tilt_series][
                    "em-tomo-preprocess"
                ],
                "mc_uuid": environment.movies[
                    file_transferred_to
                ].motion_correction_uuid,
            }
            requests.post(preproc_url, json=preproc_data)
        elif environment:
            preproc_url = f"{str(environment.url.geturl())}/visits/{environment.visit}/tomography_preprocess"
            pfi = ProcessFileIncomplete(
                dest=file_transferred_to,
                source=environment.source,
                image_number=environment.movies[file_transferred_to].movie_number,
                mc_uuid=environment.movies[file_transferred_to].motion_correction_uuid,
                tag=tilt_series,
            )
            if (
                environment.autoproc_program_ids is None
                or environment.processing_job_ids is None
            ) or (
                environment.autoproc_program_ids.get(tilt_series) is None
                or environment.processing_job_ids.get(tilt_series) is None
            ):
                if self._preprocessing_triggers.get(tilt_series):
                    self._preprocessing_triggers[tilt_series].append(
                        (
                            preproc_url,
                            pfi,
                            environment,
                        )
                    )
                else:
                    self._preprocessing_triggers[tilt_series] = [
                        (
                            preproc_url,
                            pfi,
                            environment,
                        )
                    ]

        if self._last_transferred_file:
            last_tilt_series = extract_tilt_tag(
                self._last_transferred_file
            ) + extract_tilt_series(self._last_transferred_file)
            last_tilt_angle = extract_tilt_angle(self._last_transferred_file)
            self._last_transferred_file = file_path
            if last_tilt_series != tilt_series and last_tilt_angle != tilt_angle:
                newly_completed_series = []
                if self._tilt_series:
                    tilt_series_size = max(len(ts) for ts in self._tilt_series.values())
                else:
                    tilt_series_size = 0
                this_tilt_series_size = len(self._tilt_series[tilt_series])
                if this_tilt_series_size >= tilt_series_size:
                    self._completed_tilt_series.append(tilt_series)
                    newly_completed_series.append(tilt_series)
                for ts, ta in self._tilt_series.items():
                    if (
                        len(ta) >= tilt_series_size
                        and ts not in self._completed_tilt_series
                    ):
                        newly_completed_series.append(ts)
                        self._completed_tilt_series.append(ts)
                        if environment:
                            file_tilt_list = []
                            movie: str
                            angle: str
                            for movie, angle in environment.tilt_angles[ts]:
                                if environment.motion_corrected_movies.get(Path(movie)):
                                    file_tilt_list.append(
                                        [
                                            str(
                                                environment.motion_corrected_movies[
                                                    Path(movie)
                                                ][0]
                                            ),
                                            angle,
                                            str(
                                                environment.motion_corrected_movies[
                                                    Path(movie)
                                                ][1]
                                            ),
                                        ]
                                    )
                                if environment.motion_corrected_movies.get(
                                    file_transferred_to
                                ):
                                    self._check_for_alignment(
                                        file_transferred_to,
                                        Path(
                                            environment.motion_corrected_movies[  # key error PosixPath
                                                file_transferred_to
                                            ][
                                                0
                                            ]
                                        ),
                                        environment.url.geturl(),
                                        environment.data_collection_ids[ts],
                                        environment.processing_job_ids[ts][
                                            "em-tomo-align"
                                        ],
                                        environment.autoproc_program_ids[ts][
                                            "em-tomo-align"
                                        ],
                                        int(
                                            environment.motion_corrected_movies[
                                                file_transferred_to
                                            ][1]
                                        ),
                                        file_tilt_list,
                                    )
                if newly_completed_series:
                    logger.info(
                        f"The following tilt series are considered complete: {newly_completed_series}"
                    )
                return newly_completed_series
        self._last_transferred_file = file_path
        return []

    def _add_tomo_tilt(
        self, file_path: Path, environment: MurfeyInstanceEnvironment | None = None
    ) -> List[str]:
        if "[" in file_path.name:
            return self._add_tilt(
                file_path,
                lambda x: x.name.split("_")[1],
                lambda x: x.name.split("[")[1].split("]")[0],
                lambda x: x.name.split("_")[0],
                environment=environment,
            )

        def _get_slice_index(tag: str) -> int:
            slice_index = 0
            for i, ch in enumerate(tag[::-1]):
                if not ch.isnumeric():
                    slice_index = -i
                    break
            if not slice_index:
                raise ValueError(
                    f"The file tag {tag} does not end in numeric characters or is entirely numeric: cannot parse"
                )
            return slice_index

        def _extract_tilt_series(p: Path) -> str:
            tag = p.name.split("_")[0]
            slice_index = _get_slice_index(tag)
            return tag[slice_index:]

        def _extract_tilt_tag(p: Path) -> str:
            tag = p.name.split("_")[0]
            slice_index = _get_slice_index(tag)
            return tag[:slice_index]

        def _extract_tilt_angle(p: Path) -> str:
            _split = p.name.split("_")[2].split(".")
            return ".".join(_split[:-1])

        return self._add_tilt(
            file_path,
            _extract_tilt_series,
            _extract_tilt_angle,
            _extract_tilt_tag,
            environment=environment,
        )

    def _add_serialem_tilt(
        self, file_path: Path, environment: MurfeyInstanceEnvironment | None = None
    ) -> List[str]:
        delimiters = ("_", "-")
        for d in delimiters:
            if file_path.name.count(d) > 1:
                delimiter = d
                break
        else:
            delimiter = delimiters[0]

        def _extract_tilt_series(p: Path) -> str:
            split = p.name.split(delimiter)
            for s in split:
                if s.isdigit():
                    return s
            raise ValueError(
                f"No digits found in {p.name} after splitting on {delimiter}"
            )

        return self._add_tilt(
            file_path,
            _extract_tilt_series,
            lambda x: ".".join(x.name.split(delimiter)[-1].split(".")[:-1]),
            lambda x: "",
            environment=environment,
        )

    def post_transfer(
        self,
        transferred_file: Path,
        role: str = "",
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ) -> List[str]:
        data_suffixes = (".mrc", ".tiff", ".tif", ".eer")
        completed_tilts = []
        if (
            role == "detector"
            and transferred_file.suffix in data_suffixes
            and "gain" not in transferred_file.name
        ):
            if self._acquisition_software == "tomo":
                completed_tilts = self._add_tomo_tilt(
                    transferred_file, environment=environment
                )
            elif self._acquisition_software == "serialem":
                completed_tilts = self._add_serialem_tilt(
                    transferred_file, environment=environment
                )
        return completed_tilts

    def post_first_transfer(
        self,
        transferred_file: Path,
        role: str = "",
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ):
        self.post_transfer(transferred_file, role=role, environment=environment)

    def gather_metadata(self, metadata_file: Path) -> OrderedDict:
        if metadata_file.suffix not in (".mdoc", ".xml"):
            raise ValueError(
                f"Tomography gather_metadata method expected xml or mdoc file not {metadata_file.name}"
            )
        if not metadata_file.is_file():
            logger.debug(f"Metadata file {metadata_file} not found")
            return OrderedDict({})
        if metadata_file.suffix == ".xml":
            with open(metadata_file, "r") as xml:
                try:
                    for_parsing = xml.read()
                except Exception:
                    logger.warning(f"Failed to parse file {metadata_file}")
                    return OrderedDict({})
                data = xmltodict.parse(for_parsing)
            metadata: OrderedDict = OrderedDict({})
            metadata["experiment_type"] = TUIFormValue("tomography")
            metadata["voltage"] = TUIFormValue(300)
            metadata["image_size_x"] = TUIFormValue(
                data["Acquisition"]["Info"]["ImageSize"]["Width"]
            )
            metadata["image_size_y"] = TUIFormValue(
                data["Acquisition"]["Info"]["ImageSize"]["Height"]
            )
            metadata["pixel_size_on_image"] = TUIFormValue(
                float(data["Acquisition"]["Info"]["SensorPixelSize"]["Height"])
            )
            metadata["gain_ref"] = TUIFormValue(None, top=True)
            metadata["dose_per_frame"] = TUIFormValue(
                None, top=True, colour="dark_orange"
            )
            metadata.move_to_end("gain_ref", last=False)
            metadata.move_to_end("dose_per_frame", last=False)
            # logger.info(f"Metadata extracted from {metadata_file}: {metadata}")
            return metadata
        with open(metadata_file, "r") as md:
            mdoc_data = get_global_data(md)
        if not mdoc_data:
            return OrderedDict({})
        mdoc_metadata: OrderedDict = OrderedDict({})
        mdoc_metadata["experiment_type"] = TUIFormValue("tomography")
        mdoc_metadata["voltage"] = TUIFormValue(float(mdoc_data["Voltage"]))
        mdoc_metadata["image_size_x"] = TUIFormValue(int(mdoc_data["ImageSize"][0]))
        mdoc_metadata["image_size_y"] = TUIFormValue(int(mdoc_data["ImageSize"][1]))
        mdoc_metadata["pixel_size_on_image"] = TUIFormValue(
            float(mdoc_data["PixelSpacing"]) * 1e-10
        )
        mdoc_metadata["dose_per_frame"] = TUIFormValue(
            None, top=True, colour="dark_orange"
        )
        mdoc_metadata.move_to_end("dose_per_frame", last=False)
        # logger.info(f"Metadata extracted from {metadata_file}")
        return mdoc_metadata



from __future__ import annotations

from pathlib import Path


def determine_gain_ref(gain_ref_dir: Path) -> Path:
    candidates = list(gain_ref_dir.glob("*"))
    candidates = sorted(candidates, key=lambda x: x.stat().st_mtime, reverse=True)
    viable_candidates = candidates[:3]
    viable_candidates = sorted(
        viable_candidates, key=lambda x: x.stat().st_size, reverse=True
    )
    if viable_candidates:
        return viable_candidates[0]
    raise RuntimeError(f"Cannot identify a gain reference candidate in {gain_ref_dir}")



from __future__ import annotations

from datetime import datetime
from typing import TextIO


def _basic_parse(line: str) -> dict:
    kv = [p.strip() for p in line.split("=")]
    if kv[0] == "DateTime":
        return {kv[0]: datetime.strptime(kv[1], "%d-%b-%y %H:%M:%S")}
    return {kv[0]: tuple(kv[1].split()) if " " in kv[1] else kv[1]}


def get_block(mdocfile: TextIO) -> dict:
    while line := mdocfile.readline():
        if line.startswith("[ZValue"):
            break
    else:
        return {}
    as_dict = {}
    while line := mdocfile.readline():
        if line.replace(" ", "") == "\n":
            break
        as_dict.update(_basic_parse(line))
    return as_dict


def get_global_data(mdocfile: TextIO) -> dict:
    as_dict = {}
    while line := mdocfile.readline():
        if line.startswith("[") or line.replace(" ", "") == "\n":
            break
        as_dict.update(_basic_parse(line))
    return as_dict



from __future__ import annotations

import datetime
import logging

import ispyb

# import ispyb.sqlalchemy
import sqlalchemy.orm
import workflows.transport
from fastapi import Depends
from ispyb.sqlalchemy import (
    AutoProcProgram,
    BLSession,
    DataCollection,
    DataCollectionGroup,
    ProcessingJob,
    Proposal,
    url,
)

from murfey.util.models import Visit

# _BLSession = ispyb.sqlalchemy.BLSession
# _Proposal = ispyb.sqlalchemy.Proposal
# _DataCollection = ispyb.sqlalchemy.DataCollection
# _ProcessingJob = ispyb.sqlalchemy.ProcessingJob
# _DataCollectionGroup = ispyb.sqlalchemy.DataCollectionGroup
# _AutoProcProgram = ispyb.sqlalchemy.AutoProcProgram

log = logging.getLogger("murfey.server.ispyb")

Session = sqlalchemy.orm.sessionmaker(
    bind=sqlalchemy.create_engine(url(), connect_args={"use_pure": True})
)


class TransportManager:
    def __init__(self, transport_type):
        self.transport = workflows.transport.lookup(transport_type)()
        self.transport.connect()
        self.ispyb = ispyb.open()

    def do_insert_data_collection_group(
        self, record: DataCollectionGroup, message=None, **kwargs
    ):
        dcgparams = self.ispyb.em_acquisition.get_data_collection_group_params()
        dcgparams["parentid"] = record.sessionId
        dcgparams["experimenttype"] = record.experimentType
        try:
            data_collection_group_id = (
                self.ispyb.em_acquisition.upsert_data_collection_group(
                    list(dcgparams.values())
                )
            )
            log.info(f"Created DataCollectionGroup {data_collection_group_id}")
            return {"success": True, "return_value": data_collection_group_id}
        except ispyb.ISPyBException as e:
            log.error(
                "Inserting Data Collection Group entry caused exception '%s'.",
                e,
                exc_info=True,
            )
        return False

    def do_insert_data_collection(self, record: DataCollection, message=None, **kwargs):
        dc_params = self.ispyb.em_acquisition.get_data_collection_params()
        dc_params["parentid"] = record.dataCollectionGroupId
        dc_params["visitid"] = record.SESSIONID
        dc_params["comments"] = "Created for Murfey"
        dc_params["imgdir"] = record.imageDirectory
        dc_params["imgsuffix"] = record.imageSuffix
        dc_params["voltage"] = record.voltage
        try:
            data_collection_id = self.ispyb.em_acquisition.upsert_data_collection(
                list(dc_params.values())
            )
            log.info(f"Created DataCollection {data_collection_id}")
            return {"success": True, "return_value": data_collection_id}
        except ispyb.ISPyBException as e:
            log.error(
                "Inserting Data Collection entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_create_ispyb_job(self, record: ProcessingJob, rw=None, **kwargs):
        dcid = record.dataCollectionId
        if not dcid:
            log.error("Can not create job: DCID not specified")
            return False

        jp = self.ispyb.mx_processing.get_job_params()
        jp["automatic"] = record.automatic
        jp["comments"] = record.comments
        jp["datacollectionid"] = dcid
        jp["display_name"] = record.displayName
        jp["recipe"] = record.recipe
        log.info("Creating database entries...")
        try:
            jobid = self.ispyb.mx_processing.upsert_job(list(jp.values()))
            log.info(f"All done. Processing job {jobid} created")
            return {"success": True, "return_value": jobid}
        except ispyb.ISPyBException as e:
            log.error(
                "Inserting Processing Job entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_update_processing_status(self, record: AutoProcProgram, **kwargs):
        ppid = record.autoProcProgramId
        message = record.processingMessage
        status = record.processingStatus
        try:
            result = self.ispyb.mx_processing.upsert_program_ex(
                program_id=ppid,
                status={"success": 1, "failure": 0}.get(status),
                time_start=record.processingStartTime,
                time_update=record.processingEndTime,
                message=message,
                job_id=record.processingJobId,
            )
            log.info(
                f"Updating program {ppid} with status {message!r}",
            )
            return {"success": True, "return_value": result}
        except ispyb.ISPyBException as e:
            log.error(
                "Updating program %s status: '%s' caused exception '%s'.",
                ppid,
                message,
                e,
                exc_info=True,
            )
            return False


def _get_session() -> sqlalchemy.orm.Session:
    db = Session()
    try:
        yield db
    finally:
        db.close()


DB = Depends(_get_session)
# Shortcut to access the database in a FastAPI endpoint


def get_session_id(
    microscope: str,
    proposal_code: str,
    proposal_number: str,
    visit_number: str,
    db: sqlalchemy.orm.Session,
) -> int:
    query = (
        db.query(BLSession)
        .join(Proposal)
        .filter(
            BLSession.proposalId == Proposal.proposalId,
            BLSession.beamLineName == microscope,
            Proposal.proposalCode == proposal_code,
            Proposal.proposalNumber == proposal_number,
            BLSession.visit_number == visit_number,
        )
        .add_columns(BLSession.sessionId)
        .all()
    )
    return query[0][1]


def get_all_ongoing_visits(microscope: str, db: sqlalchemy.orm.Session) -> list[Visit]:
    query = (
        db.query(BLSession)
        .join(Proposal)
        .filter(
            BLSession.proposalId == Proposal.proposalId,
            BLSession.beamLineName == microscope,
            BLSession.endDate > datetime.datetime.now(),
            BLSession.startDate < datetime.datetime.now(),
        )
        .add_columns(
            BLSession.startDate,
            BLSession.endDate,
            BLSession.sessionId,
            Proposal.proposalCode,
            Proposal.proposalNumber,
            BLSession.visit_number,
            Proposal.title,
        )
        .all()
    )
    return [
        Visit(
            start=row.startDate,
            end=row.endDate,
            session_id=row.sessionId,
            name=f"{row.proposalCode}{row.proposalNumber}-{row.visit_number}",
            proposal_title=row.title,
            beamline=microscope,
        )
        for row in query
    ]


def get_data_collection_group_ids(session_id):
    query = (
        Session()
        .query(DataCollectionGroup)
        .filter(
            DataCollectionGroup.sessionId == session_id,
        )
        .all()
    )
    dcgids = [row.dataCollectionGroupId for row in query]
    return dcgids
